@@ -39,6 +39,55 @@
   uniroot(expectation, interval = integral_interval, delta)$root
 }
 
+## Determine a breakdown point with stable numerical properties of the M-scale
+## with Tukey's bisquare rho function.
+##
+## The M-scale objective (and hence the S-loss) can have unbounded or very high
+## 1st derivative. This can lead to numerical instability of the algorithms and
+## in turn excessive computation time.
+## This function chooses the breakdown point with lowest upper bound of the 1st
+## derivative from a range of bdp's in the vicinity of the desired bdp.
+##
+## @param n number of observations in the sample
+## @param desired_bdp the desired breakdown point (between 0.05 and 0.5)
+## @param tolerance how far can the chosen bdp be away from the desired bdp.
+##                  The chosen bdp is guaranteed to be in the range given by `interval`.
+## @param interval restrict the chosen bdp to this interval.
+## @param precision granularity of the grid of considered bdp's.
+.find_stable_bdb_bisquare <- function (n, desired_bdp, tolerance = 0.01, precision = 1e-4,
+                                       interval = c(0.05, 0.5)) {
+  numeric_tol <- sqrt(.Machine$double.eps)
+
+  bdp_range <- seq(from = max(desired_bdp - tolerance, interval[[1L]]),
+                   to = min(desired_bdp + tolerance, interval[[2L]]),
+                   by = precision)
+
+  # Filter bdp's where the 1st derivative is unbounded
+  bdp_range <- bdp_range[abs(bdp_range * n - floor(bdp_range * n)) > numeric_tol]
+
+  # Determine an upper bound for the 1st derivative of the M-scale objective function
+  first_deriv_bound <- vapply(bdp_range, FUN.VALUE = numeric(1L), FUN = function (bdp) {
+    thresh <- tryCatch(uniroot(f = function (t) {
+      up <- n * (1 - bdp) / (1 - t)
+      up - floor(up) - n * t / (1 - t)
+    }, interval = c(0, 0.5),  extendInt = 'downX', tol = numeric_tol)$root,
+    error = function (e) {
+      return(NA_real_)
+    })
+
+    1 / sqrt(1 - (1 - thresh)^(1/3))
+  })
+  good_bounds <- which(is.finite(first_deriv_bound))
+
+  if (length(good_bounds) == 0L) {
+    warning(paste("The chosen breakdown point may lead to numerical instability and",
+                  "excessive computation time.",
+                  "Consider changing the breakdown point via argument `bdp`."))
+    return(desired_bdp)
+  }
+  bdp_range[[which.min(first_deriv_bound)]]
+}
+
 ## Approximate Value Matching
 ##
 ## @param x,table see [base::match] for details.
@@ -46,7 +95,10 @@
 ## @return a vector the same length as `x` with integers giving the position in
 ##         `table` of the first match if there is a match, or `NA_integer_`
 ##         otherwise.
-.approx_match <- function(x, table, eps = min(sqrt(.Machine$double.eps), 0.5 * min(x, table))) {
+.approx_match <- function(x, table, eps) {
+  if (missing(eps)) {
+    eps <- max(.Machine$double.eps, min(sqrt(.Machine$double.eps), 0.5 * min(x, table)))
+  }
   .Call(C_approx_match, as.numeric(x), as.numeric(table), as.numeric(eps[[1L]]))
 }
 
@@ -84,9 +136,11 @@ extract_metric <- function (metrics, attr, node) {
 ## @param std_data standardized full data set (standardized by `.standardize_data`)
 ## @param cv_k number of folds per CV split
 ## @param cv_repl number of CV replications.
-## @param cv_est_fun function taking the standardized training set and the indices of the left-out observations and
-##                   returns a list of estimates. The function always needs to return the same number of estimates!
-## @param metric function taking a vector of prediction errors and returning the scale of the prediction error.
+## @param cv_est_fun function taking the standardized training set and the indices of
+##    the left-out observations and returns a list of estimates.
+##    The function always needs to return the same number of estimates!
+## @param metric function taking a vector of prediction errors and returning the scale
+##    of the prediction error.
 #' @importFrom Matrix drop
 #' @importFrom rlang abort
 .run_replicated_cv <- function (std_data, cv_k, cv_repl, cv_est_fun, metric, par_cluster = NULL) {
@@ -119,22 +173,25 @@ extract_metric <- function (metrics, attr, node) {
   }, est_fun = est_fun)
 
   predictions_all <- split(predictions_all, rep(seq_len(cv_repl), each = cv_k))
-  prediction_metrics <- mapply(predictions_all, test_segments_list, FUN = function (predictions, test_inds) {
-    obs_order <- sort.list(unlist(test_inds, recursive = FALSE, use.names = FALSE))
-    ordered_predictions <- do.call(rbind, predictions)[obs_order, ]
+  prediction_metrics <- mapply(
+    predictions_all, test_segments_list,
+    FUN = function (predictions, test_inds) {
+      obs_order <- sort.list(unlist(test_inds, recursive = FALSE, use.names = FALSE))
+      ordered_predictions <- do.call(rbind, predictions)[obs_order, ]
 
-    if (call_with_errors) {
-      apply(ordered_predictions - std_data$y, 2, metric)
-    } else {
-      apply(ordered_predictions, 2, metric, std_data$y)
-    }
-  })
+      if (call_with_errors) {
+        apply(ordered_predictions - std_data$y, 2, metric)
+      } else {
+        apply(ordered_predictions, 2, metric, std_data$y)
+      }
+    })
   matrix(unlist(prediction_metrics, recursive = FALSE, use.names = FALSE), ncol = cv_repl)
 }
 
 ## Standardize data
 ##
-## @param x predictor matrix. Can also be a list with components `x` and `y`, in which case `y` is ignored.
+## @param x predictor matrix. Can also be a list with components `x` and `y`,
+##    in which case `y` is ignored.
 ## @param y response vector.
 ## @param intercept is an intercept included (i.e., should `y` be centered?)
 ## @param standardize standardize or not.
@@ -146,8 +203,9 @@ extract_metric <- function (metrics, attr, node) {
 #' @importFrom methods is
 #' @importFrom rlang abort
 #' @importFrom stats sd
-.standardize_data <- function (x, y, intercept, standardize, robust, sparse, mscale_opts, location_rho = 'bisquare',
-                               location_cc = 4.5, target_scale_x = NULL, ...) {
+.standardize_data <- function (x, y, intercept, standardize, robust, sparse, mscale_opts,
+                               location_rho = 'bisquare', location_cc = 4.5,
+                               target_scale_x = NULL, ...) {
   if (is.list(x) && !is.null(x$x) && !is.null(x$y)) {
     y <- x$y
     x <- x$x
@@ -181,7 +239,8 @@ extract_metric <- function (metrics, attr, node) {
       apply(ret_list$x, 2, sd)
     } else {
       locscale <- apply(ret_list$x, 2, function (xj) {
-        mlocscale(xj, location_rho = location_rho, location_cc = location_cc, opts = mscale_opts, ...)
+        mlocscale(xj, location_rho = location_rho, location_cc = location_cc,
+                  opts = mscale_opts, ...)
       })
       if (isTRUE(intercept)) {
         # Re-center the predictors with the updated centers
@@ -217,12 +276,25 @@ extract_metric <- function (metrics, attr, node) {
 
     if (isTRUE(standardize == 'cv_only')) {
       # In case of "CV only" standardization, match the original scaling
-      .standardize_data(x, y, intercept = intercept, standardize = TRUE, robust = robust, sparse = sparse,
-                        location_rho = location_rho, location_cc = location_cc, mscale_opts = mscale_opts,
-                        target_scale_x = ret_list$scale_x, ... = ...)
+      .standardize_data(x, y,
+                        intercept = intercept,
+                        standardize = TRUE, robust = robust,
+                        sparse = sparse,
+                        location_rho = location_rho,
+                        location_cc = location_cc,
+                        mscale_opts = mscale_opts,
+                        target_scale_x = ret_list$scale_x,
+                        ... = ...)
     } else {
-      .standardize_data(x, y, intercept = intercept, standardize = standardize, robust = robust, sparse = sparse,
-                        location_rho = location_rho, location_cc = location_cc, mscale_opts = mscale_opts, ... = ...)
+      .standardize_data(x, y,
+                        intercept = intercept,
+                        standardize = standardize,
+                        robust = robust,
+                        sparse = sparse,
+                        location_rho = location_rho,
+                        location_cc = location_cc,
+                        mscale_opts = mscale_opts,
+                        ... = ...)
     }
   }
 
@@ -260,7 +332,8 @@ extract_metric <- function (metrics, attr, node) {
       }
       if (isTRUE(intercept)) {
         # Recreate intercept
-        coef_obj$intercept <- coef_obj$intercept + ret_list$muy - sum(ret_list$mux[coef_obj$beta@i] * coef_obj$beta@x)
+        coef_obj$intercept <- coef_obj$intercept + ret_list$muy -
+          sum(ret_list$mux[coef_obj$beta@i] * coef_obj$beta@x)
       }
       return(coef_obj)
     }
@@ -315,7 +388,7 @@ extract_metric <- function (metrics, attr, node) {
 
   # "ignore" solutions after which the prediction performance comes back down
   best_1se <- if (any(diff(candidates) > 1)) {
-    min(candidates[seq_len(max(which(diff(candidates) > 1)))])
+    min(candidates[-seq_len(max(which(diff(candidates) > 1)))])
   } else {
     min(candidates)
   }
@@ -385,11 +458,12 @@ extract_metric <- function (metrics, attr, node) {
   }
 }
 
-.prepare_penalty_loadings <- function (penalty_loadings, x, alpha, sparse, stop_all_infinite = FALSE) {
+.prepare_penalty_loadings <- function (penalty_loadings, x, alpha, sparse,
+                                       stop_all_infinite = FALSE) {
   orig_p <- ncol(x)
   restore_coef_length <- function (coef) coef
 
-  if(alpha < .Machine$double.eps) {
+  if(any(alpha < .Machine$double.eps)) {
     abort("Non-empty `penalty_loadings` only supported for `alpha` > 0.")
   } else if (length(penalty_loadings) != orig_p) {
     abort("`penalty_loadings` has different number of elements than `x` columns.")
@@ -492,3 +566,29 @@ extract_metric <- function (metrics, attr, node) {
     })
   }
 }
+
+## Parse strings of the form *min*, *se*, or *{m}-se*.
+#' @importFrom rlang abort
+.parse_se_string <- function (x, only_fact = FALSE) {
+  x <- .as(x[[1L]], 'character')
+  xlen <- nchar(x)
+  se_fact <- 1
+  se_str <- if (identical('-se', substr(x, xlen - 2L, xlen))) {
+    se_fact <- as.numeric(substr(x, 0L, xlen - 3L))
+    if (anyNA(se_fact)) {
+      abort(sprintf("Cannot parse standard error string '%s'.", x))
+    }
+    'se'
+  } else {
+    match.arg(x, c('min', 'se'))
+  }
+  if (identical(se_str, 'min')) {
+    se_fact <- 0
+  }
+
+  if (isTRUE(only_fact)) {
+    se_fact
+  } else {
+    list(se_type = se_str, se_fact = se_fact)
+  }
+}
