@@ -26,12 +26,13 @@ tau_size <- function (x) {
 #'
 #' @param x numeric values. Missing values are verbosely ignored.
 #' @param bdp desired breakdown point (between 0 and 0.5).
-#' @param cc cutoff value for the bisquare rho function. By default, chosen to yield a consistent estimate for the
-#'    Normal distribution.
-#' @param opts a list of options for the M-scale estimation algorithm, see [mscale_algorithm_options()]
-#'    for details.
+#' @param cc cutoff value for the bisquare rho function.
+#'    By default, chosen to yield a consistent estimate for the Normal distribution.
+#' @param opts a list of options for the M-scale estimation algorithm,
+#'    see [mscale_algorithm_options()] for details.
 #' @param delta deprecated. Use `bpd` instead.
-#' @param rho,eps,maxit deprecated. Instead set control options for the algorithm with the `opts` arguments.
+#' @param rho,eps,maxit deprecated. Instead set control options for the algorithm
+#'    with the `opts` arguments.
 #' @return the M-estimate of scale.
 #'
 #' @family functions to compute robust estimates of location and scale
@@ -42,8 +43,8 @@ tau_size <- function (x) {
 #' @importFrom rlang warn
 #' @importFrom stats na.omit
 mscale <- function (x, bdp = 0.25, cc = consistency_const(bdp, 'bisquare'),
-                    opts = mscale_algorithm_options(), delta = deprecated(), rho = deprecated(), eps = deprecated(),
-                    maxit = deprecated()) {
+                    opts = mscale_algorithm_options(), delta = deprecated(),
+                    rho = deprecated(), eps = deprecated(), maxit = deprecated()) {
   if (is_present(delta)) {
     deprecate_warn('2.0.0', 'mscale(delta=)', 'mscale(bdp=)')
     bdp <- delta
@@ -105,7 +106,7 @@ mloc <- function (x, scale, rho, cc, opts = mscale_algorithm_options()) {
     scale <- mad(x)
   }
 
-  if (scale < .Machine$double.eps) {
+  if (!isTRUE(scale > .Machine$double.eps)) {
     warn("Cannot compute M-estimate of location for values with scale of 0.")
     return(NA_real_)
   }
@@ -124,9 +125,11 @@ mloc <- function (x, scale, rho, cc, opts = mscale_algorithm_options()) {
 #'
 #' @param x numeric values. Missing values are verbosely ignored.
 #' @param bdp desired breakdown point (between 0 and 0.5).
-#' @param scale_cc cutoff value for the bisquare \eqn{\rho} function for computing the scale estimate. By default, chosen
-#'    to yield a consistent estimate for normally distributed values.
-#' @param location_rho,location_cc \eqn{\rho} function and cutoff value for computing the location estimate.
+#' @param scale_cc cutoff value for the bisquare \eqn{\rho} function for computing the
+#'    scale estimate.
+#'    By default, chosen to yield a consistent estimate for normally distributed values.
+#' @param location_rho,location_cc \eqn{\rho} function and cutoff value for computing
+#'    the location estimate.
 #'    See [rho_function()] for a list of available \eqn{\rho} functions.
 #' @param opts a list of options for the M-estimating equation,
 #'    see [mscale_algorithm_options()] for details.
@@ -196,24 +199,34 @@ rho_function <- function (rho) {
 #' Multiple starting points can be created by combining starting points via
 #' `c(starting_point_1, starting_point_2, ...)`.
 #'
-#' A starting points can either be *shared*, i.e., used for every penalization level PENSE estimates are computed for,
-#' or *specific* to one penalization level.
-#' To create a specific starting point, provide the penalization level as `lambda`. If `lambda` is missing,
-#' a shared starting point is created.
-#' Shared and specific starting points can all be combined into a single list of starting points, with
-#' [pense()] handling them correctly.
-#' Note that specific starting points will lead to the `lambda` value being added to the grid of penalization levels.
+#' A starting points can either be *shared*, i.e., used for every penalization level PENSE
+#' estimates are computed for, or *specific* to one penalization level.
+#' To create a specific starting point, provide the penalization parameters `lambda` and `alpha`.
+#' If `lambda` or `alpha` are missing, a shared starting point is created.
+#' Shared and specific starting points can all be combined into a single list of starting points,
+#' with [pense()] handling them correctly.
+#' Note that specific starting points will lead to the `lambda` value being added to the
+#' grid of penalization levels.
 #' See [pense()] for details.
 #'
-#' Starting points computed via [enpy_initial_estimates()] are by default *shared* starting points but can
-#' be transformed to *specific* starting points via `enpy_starting_point(..., specific = TRUE)`.
+#' Starting points computed via [enpy_initial_estimates()] are by default *shared* starting points
+#' but can be transformed to *specific* starting points via
+#' `as_starting_point(..., specific = TRUE)`.
 #'
-#' @param beta beta coefficients at the starting point. Can be a numeric vector, a sparse vector of class
-#'    [dsparseVector][Matrix::sparseVector-class], or a sparse matrix of class
-#'    [dgCMatrix][Matrix::CsparseMatrix-class] with a single column.
+#' @param beta beta coefficients at the starting point. Can be a numeric vector,
+#'    a sparse vector of class [dsparseVector][Matrix::sparseVector-class],
+#'    or a sparse matrix of class [dgCMatrix][Matrix::CsparseMatrix-class] with a single column.
 #' @param intercept intercept coefficient at the starting point.
-#' @param lambda optional penalization level. If missing, the starting point is used at *every* penalization level,
-#'    otherwise, only at this penalization level.
+#' @param alpha optional value(s) for the `alpha` hyper-parameter.
+#'    For `starting_point()`, makes the starting point specific to the given `alpha` value.
+#'    For `as_starting_point()`, only estimates with matching `alpha` values are extracted.
+#'    Values not present in `object` are ignored with a warning.
+#' @param lambda optional penalization level(s).
+#'    For `starting_point()`, makes the starting point specific to the given `alpha` value.
+#'    For `as_starting_point()`, only estimates with matching `alpha` and `lambda` values
+#'    are extracted.
+#'    If `lambda` is given, `alpha` must be given and be a single number.
+#'    Penalization levels not present in `object` are ignored with a warning.
 #' @return an object of type `starting_points` to be used as starting point for [pense()].
 #'
 #' @family functions for initial estimates
@@ -223,12 +236,13 @@ rho_function <- function (rho) {
 #' @importFrom methods is
 #' @importFrom Matrix sparseVector
 #' @importFrom rlang abort
-starting_point <- function (beta, intercept, lambda) {
+starting_point <- function (beta, intercept, lambda, alpha) {
   sp <- list(intercept = .as(intercept[[1L]], 'numeric'))
-  if (missing(lambda)) {
+  if (missing(lambda) && missing(alpha)) {
     class(sp) <- c('shared_starting_point', 'starting_point')
   } else {
-    sp$lambda <- .as(lambda, 'numeric')
+    sp$lambda <- .as(lambda[[1L]], 'numeric')
+    sp$alpha <- .as(alpha[[1L]], 'numeric')
     class(sp) <- c('specific_starting_point', 'starting_point')
   }
   if (is(beta, 'dgCMatrix') && ncol(beta) == 1L) {
@@ -245,8 +259,9 @@ starting_point <- function (beta, intercept, lambda) {
 
 #' @rdname starting_point
 #' @param object an object with estimates to use as starting points.
-#' @param specific whether the estimates should be used as starting points only at the penalization level they
-#'   are computed for. Defaults to using the estimates as starting points for all penalization levels.
+#' @param specific whether the estimates should be used as starting points only at
+#'   the penalization level they are computed for.
+#'   Defaults to using the estimates as starting points for all penalization levels.
 #' @param ... further arguments passed to or from other methods.
 #' @export
 as_starting_point <- function (object, specific = FALSE, ...) {
@@ -265,23 +280,43 @@ as_starting_point.enpy_starting_points <- function (object, specific = FALSE, ..
 }
 
 #' @rdname starting_point
-#' @param lambda optional penalization level(s) to extract from `object`. Penalization levels not present in `object`
-#'    are ignored with a warning.
 #' @export
 #' @importFrom rlang warn abort
-as_starting_point.pense_fit <- function (object, specific = FALSE, lambda, ...) {
+as_starting_point.pense_fit <- function (object, specific = FALSE, alpha, lambda, ...) {
+  alpha <- if (missing(alpha) || is.null(alpha)) {
+    orig_alpha_length <- length(object$alpha)
+    object$alpha
+  } else {
+    orig_alpha_length <- length(alpha)
+    object$alpha[na.omit(.approx_match(.as(alpha, 'numeric'), object$alpha))]
+  }
+  if (length(alpha) == 0L) {
+    abort("No requested `alpha` values are not available in `object`.")
+  } else if (length(alpha) < orig_alpha_length) {
+    warn("Some requested `alpha` values are not available in `object`.")
+  }
+
+  alpha_indices <- which(vapply(object$estimates, FUN.VALUE = logical(1L), FUN = function (est) {
+    any((est$alpha - alpha)^2 < .Machine$double.eps)
+  }))
+
   lambda_indices <- if (missing(lambda)) {
-    seq_along(object$estimates)
+    alpha_indices
   } else if (is.numeric(lambda)) {
-    lambda_indices <- .approx_match(lambda, object$lambda)
+    if (length(alpha) > 1L) {
+      abort("If `lambda` is given `alpha` must be a single number.")
+    }
+    ai_lambdas <- vapply(object$estimates[alpha_indices], FUN.VALUE = numeric(1L),
+                         FUN = `[[`, 'lambda')
+    lambda_indices <- .approx_match(lambda, ai_lambdas)
     bad_lambda_indices <- which(is.na(lambda_indices))
     if (length(bad_lambda_indices) == length(lambda)) {
       abort("No penalization level requested with `lambda` is available in `object`.")
     } else if (length(bad_lambda_indices) > 0L) {
       warn("Some penalization levels requested with `lambda` are not available in `object`.")
-      lambda_indices[-bad_lambda_indices]
+      alpha_indices[lambda_indices[-bad_lambda_indices]]
     } else {
-      lambda_indices
+      alpha_indices[lambda_indices]
     }
   } else {
     abort("`lambda` must be missing or a numeric vector.")
@@ -293,34 +328,60 @@ as_starting_point.pense_fit <- function (object, specific = FALSE, lambda, ...)
     c('shared_starting_point', 'starting_point')
   }
 
-  structure(lapply(object$estimates[lambda_indices], structure, class = class_list), class = 'starting_points')
+  structure(lapply(object$estimates[lambda_indices], structure, class = class_list),
+            class = 'starting_points')
 }
 
-
 #' @rdname starting_point
-#' @param lambda optionally either a string specifying which penalty level to use (`"min"` or `"se"`) or a numeric
-#'    vector of the penalty levels to extract from `object`. Penalization levels not present in `object`
-#'    are ignored with a warning. If `NULL`, all estimates in `object` are extracted.
+#' @param alpha optional value for the `alpha` hyper-parameter.
+#'    If given, only estimates with matching `alpha` values are extracted.
+#'    Values not present in `object` are ignored with a warning.
+#' @param lambda optionally either a string specifying which penalty level to use
+#'    (`"min"` or `"se"`) or a numeric vector of the penalty levels to extract from `object`.
+#'    Penalization levels not present in `object` are ignored with a warning.
+#'    If `NULL`, all estimates in `object` are extracted.
+#'    If a numeric vector, `alpha` must be given and a single number.
 #' @param se_mult If `lambda = "se"`, the multiple of standard errors to tolerate.
 #'
 #' @details
-#' When creating starting points from cross-validated fits, it is possible to extract only the estimate with best
-#' CV performance (`lambda = "min"`), or the estimate with CV performance statistically indistinguishable from the
-#' best performance (`lambda = "se"`). This is determined to be the estimate with prediction performance
-#' within `se_mult * cv_se` from the best model.
+#' When creating starting points from cross-validated fits, it is possible to extract only the
+#' estimate with best CV performance (`lambda = "min"`), or the estimate with CV performance
+#' statistically indistinguishable from the best performance (`lambda = "se"`).
+#' This is determined to be the estimate with prediction performance within
+#' `se_mult * cv_se` from the best model.
 #'
 #' @export
 #' @importFrom rlang warn abort
-as_starting_point.pense_cvfit <- function (object, specific = FALSE, lambda = c('min', 'se'), se_mult = 1, ...) {
+as_starting_point.pense_cvfit <- function (object, specific = FALSE,
+                                           alpha, lambda = c('min', 'se'),
+                                           se_mult = 1, ...) {
+  alpha <- if (missing(alpha) || is.null(alpha)) {
+    orig_alpha_length <- length(object$alpha)
+    object$alpha
+  } else {
+    orig_alpha_length <- length(alpha)
+    object$alpha[na.omit(.approx_match(.as(alpha, 'numeric'), object$alpha))]
+  }
+  if (length(alpha) == 0L) {
+    abort("No requested `alpha` values are not available in `object`.")
+  } else if (length(alpha) < orig_alpha_length) {
+    warn("Some requested `alpha` values are not available in `object`.")
+  }
+
+  alpha_indices <- which(vapply(object$estimates, FUN.VALUE = logical(1L), FUN = function (est) {
+    any((est$alpha - alpha)^2 < .Machine$double.eps)
+  }))
+
   lambda_indices <- if (isFALSE(object$call$fit_all)) {
     if (is.character(lambda)) {
       lambda <- match.arg(lambda)
     }
     if (!missing(lambda) && !isTRUE(lambda == 'min')) {
-      warn(paste("`object` was created with `fit_all = FALSE`. Only the estimate at the minimum is available and will",
-                 "be used as starting point."))
+      warn(paste("`object` was created with `fit_all = FALSE`.",
+                 "Only the estimates at the minimum is available and",
+                 "will be used as starting points."))
     }
-    1L
+    alpha_indices
   } else if (is.character(lambda)) {
     lambda <- match.arg(lambda)
     se_mult <- if (lambda == 'min') {
@@ -328,21 +389,30 @@ as_starting_point.pense_cvfit <- function (object, specific = FALSE, lambda = c(
     } else {
       .as(se_mult[[1L]], 'numeric')
     }
-    se_selection <- .cv_se_selection(object$cvres$cvavg, object$cvres$cvse, se_mult)
-    which(se_selection == 'se_fact')
+
+    vapply(alpha, FUN.VALUE = integer(1L), FUN = function (al) {
+      rows <- which((object$cvres$alpha - al)^2 < .Machine$double.eps)
+      se_selection <- .cv_se_selection(object$cvres$cvavg[rows], object$cvres$cvse[rows], se_mult)
+      rows[which(se_selection == 'se_fact')]
+    })
   } else if (is.numeric(lambda)) {
-    lambda_indices <- .approx_match(lambda, object$lambda)
+    if (length(alpha) > 1L) {
+      abort("If `lambda` is numeric, `alpha` must be a single number.")
+    }
+    ai_lambdas <- vapply(object$estimates[alpha_indices], FUN.VALUE = numeric(1L),
+                         FUN = `[[`, 'lambda')
+    lambda_indices <- .approx_match(lambda, ai_lambdas)
     bad_lambda_indices <- which(is.na(lambda_indices))
     if (length(bad_lambda_indices) == length(lambda)) {
       abort("No penalization level requested with `lambda` is available in `object`.")
     } else if (length(bad_lambda_indices) > 0L) {
       warn("Some penalization levels requested with `lambda` are not available in `object`.")
-      lambda_indices[-bad_lambda_indices]
+      alpha_indices[lambda_indices[-bad_lambda_indices]]
     } else {
-      lambda_indices
+      alpha_indices[lambda_indices]
     }
   } else if (is.null(lambda)) {
-    seq_along(object$estimates)
+    alpha_indices
   } else {
     abort("`lambda` must be missing, a numeric vector, or either one of \"min\" or \"se\".")
   }
@@ -353,7 +423,8 @@ as_starting_point.pense_cvfit <- function (object, specific = FALSE, lambda = c(
     c('shared_starting_point', 'starting_point')
   }
 
-  structure(lapply(object$estimates[lambda_indices], structure, class = class_list), class = 'starting_points')
+  structure(lapply(object$estimates[lambda_indices], structure, class = class_list),
+            class = 'starting_points')
 }
 
 #' @export
@@ -361,10 +432,11 @@ as_starting_point.pense_cvfit <- function (object, specific = FALSE, lambda = c(
 #' @importFrom rlang abort
 as_starting_point.starting_point <- function (object, specific = FALSE, ...) {
   if (isTRUE(specific)) {
-    if (!is.null(object$lambda)) {
+    if (!is.null(object$lambda) && !is.null(object$alpha)) {
       class(object) <- c('specific_starting_point', 'starting_point')
     } else {
-      abort("`object` can not be made a specific starting point because it does not contain item `lambda`")
+      abort(paste("`object` can not be made a specific starting point because it does",
+                  "not contain items `lambda` and `alpha`."))
     }
   } else {
     class(object) <- c('shared_starting_point', 'starting_point')
@@ -389,7 +461,8 @@ c.starting_points <- function (...) {
     } else if (is(sp, 'starting_points')) {
       sp
     } else {
-      abort(sprintf("Do not know how to combine starting points with an object of type `%s`", class(sp)))
+      abort(sprintf("Do not know how to combine starting points with an object of type `%s`",
+                    class(sp)))
     }
   }), recursive = FALSE, use.names = FALSE), class = 'starting_points')
 }
